
package com.vizitsolutions.identitytable;

import javax.swing.ActionMap;
import javax.swing.table.DefaultTableColumnModel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JViewport;
import javax.swing.ListSelectionModel;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;

/**
 * <p>$Id$</p>
 *
 * <p>
 * This class maps a TableModel into two JTables, and places
 * them into a JScrollPane. The first JTable contains n fixed
 * columns, and represents what is commonly called a rowheader.
 * This is placed in the rowheader area of the JScrollPane. These
 * columns will not scroll horizontally. This is particularly
 * useful when you have one or more columns that identify an object
 * and several additional columns that contain attributes of the
 * object. Placing the identifying columns in these fixed columns
 * ensures that they are always visible while the attributes of the
 * object are scrolled into view.
 * </p>
 * <p>
 * The second JTable contains the remaining columns and is placed in
 * the main viewport and will scroll normally.
 * </p>
 *
 * @author Alex Kluge
 * @version $Revision$, $Date$
 */
public class IdentityTable4 extends JScrollPane
{
    // These names follow a similar pattern to those used in JTable UI delegate.
    public static final String LEFT_ARROW_ACTION_NAME  = "selectPreviousColumn";
    public static final String RIGHT_ARROW_ACTION_NAME = "selectNextColumn";
    public static final String SHIFT_TAB_ACTION_NAME   = "selectPreviousColumnCell";
    public static final String TAB_ACTION_NAME         = "selectNextColumnCell";
    
    private JTable           mainTable;
    private TableColumnModel mainColumnModel      = new DefaultTableColumnModel();
    private JTable           rowHeaderTable;
    private TableColumnModel rowHeaderColumnModel = new DefaultTableColumnModel();
    
    /**
     * Creates a new instance of IdentityTable1
     */
    public IdentityTable4(IdentityTableModel baseTableModel)
    {
        // Start out life as an empty scrollpane.
        super();
        
        // Loop over each of the fixed columns, and create columns in the row
        // header column model for them.
        TableColumn column;
        int         columnCount         = baseTableModel.getColumnCount();
        int         currentColumn;
        int         identityColumnCount = baseTableModel.getIdentityColumnCount();
        
        for (currentColumn = 0; currentColumn<identityColumnCount; currentColumn++)
        {
            column = new TableColumn(currentColumn);
            column.setHeaderValue(baseTableModel.getColumnName(currentColumn));
            
            rowHeaderColumnModel.addColumn(column);
        }
        
        // Now loop over the remaining columns, and add them to the main table's column model.
        for (currentColumn = identityColumnCount; currentColumn<columnCount; currentColumn++)
        {
            column = new TableColumn(currentColumn);
            column.setHeaderValue(baseTableModel.getColumnName(currentColumn));
            
            mainColumnModel.addColumn(column);
        }
        
        // Providing a column model here automatically sets the
        // autoCreateColumnsFromModel flag to false.
        rowHeaderTable = new JTable(baseTableModel, rowHeaderColumnModel);
        rowHeaderTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);

        mainTable      = new JTable(baseTableModel, mainColumnModel);
        mainTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        
        // Ensure that both tables have common selections.
        // Save the row selection model for later.
        ListSelectionModel rowSelectionModel = mainTable.getSelectionModel();
        rowHeaderTable.setSelectionModel(rowSelectionModel);
        rowHeaderTable.setPreferredScrollableViewportSize(rowHeaderTable.getPreferredSize());
        
        // This puts the column headers in the upper left corner above the row
        // header columns. This is the same thing that happens when a JTable configures
        // the containing JScrollPane.
        setCorner(JScrollPane.UPPER_LEFT_CORNER, rowHeaderTable.getTableHeader());
        
        JViewport rowHeaderViewPort = new JViewport();
        rowHeaderViewPort.setView(rowHeaderTable);
        setRowHeader(rowHeaderViewPort);

        // Listen for resize events on the row header, and resize the main table as appropriate.
        RowHeaderResizeListener rowHeaderResizeListener = new RowHeaderResizeListener(this,           rowHeaderViewPort,
                                                                                      rowHeaderTable, mainTable);
        rowHeaderTable.addComponentListener(rowHeaderResizeListener);

        setViewportView(mainTable);
        
       installKeybordActions(rowHeaderTable,                      mainTable,
                              rowSelectionModel,                   rowHeaderColumnModel.getSelectionModel(),
                              mainColumnModel.getSelectionModel());
    }
    
    protected void installKeybordActions(JTable             rowHeaderTable,           final JTable       mainTable,
                                         ListSelectionModel rowSelectionModel,        ListSelectionModel headerColumnSelectionModel,
                                         ListSelectionModel mainColumnSelectionModel)
    {
        ActionMap rowHeaderActionmap = rowHeaderTable.getActionMap();
        ActionMap mainActionMap      = mainTable.getActionMap();


         // Tab action handler, shift forward one cell, with wrapping
        TabKeyActionHandler tabKeyActionHandler
                    = new TabKeyActionHandler(rowHeaderTable,           mainTable,
                                              rowSelectionModel,        headerColumnSelectionModel,
                                              mainColumnSelectionModel, 1,
                                              true);
        // Put the action handler into the appropriate place for both the
        // row header and the main table.
        rowHeaderActionmap.put(TAB_ACTION_NAME, tabKeyActionHandler);
        mainActionMap.put(TAB_ACTION_NAME,      tabKeyActionHandler);

        // Right arrow action handler, shift forward one cell, no wrapping.
        tabKeyActionHandler
                    = new TabKeyActionHandler(rowHeaderTable,           mainTable,
                                              rowSelectionModel,        headerColumnSelectionModel,
                                              mainColumnSelectionModel, 1,
                                              false);

        rowHeaderActionmap.put(RIGHT_ARROW_ACTION_NAME, tabKeyActionHandler);
        mainActionMap.put(RIGHT_ARROW_ACTION_NAME,      tabKeyActionHandler);

        // Shift-Tab action handler, shift backward one cell, with wrapping
        tabKeyActionHandler
                    = new TabKeyActionHandler(rowHeaderTable,           mainTable,
                                              rowSelectionModel,        headerColumnSelectionModel,
                                              mainColumnSelectionModel, -1,
                                              true);

        rowHeaderActionmap.put(SHIFT_TAB_ACTION_NAME, tabKeyActionHandler);
        mainActionMap.put(SHIFT_TAB_ACTION_NAME,      tabKeyActionHandler);

        // Left arrow action handler, shift backward one cell, no wrapping.
        tabKeyActionHandler
                    = new TabKeyActionHandler(rowHeaderTable,           mainTable,
                                              rowSelectionModel,        headerColumnSelectionModel,
                                              mainColumnSelectionModel, -1,
                                              false);

        rowHeaderActionmap.put(LEFT_ARROW_ACTION_NAME, tabKeyActionHandler);
        mainActionMap.put(LEFT_ARROW_ACTION_NAME,      tabKeyActionHandler);
    }
    
    public TableColumnModel getMainColumnModel()
    {
        return mainColumnModel;
    }
    
    public TableColumnModel getRowHeaderColumnModel()
    {
        return rowHeaderColumnModel;
    }
}
