
package com.vizitsolutions.identitytable;

import javax.swing.ActionMap;
import javax.swing.table.DefaultTableColumnModel;
import javax.swing.InputMap;
import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JViewport;
import java.awt.event.KeyEvent;
import javax.swing.KeyStroke;
import javax.swing.ListSelectionModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;

/**
 * <p>$Id$</p>
 *
 * <p>
 * This class maps a TableModel into two JTables, and places
 * them into a JScrollPane. The first JTable contains n fixed
 * columns, and represents what is commonly called a rowheader.
 * This is placed in the rowheader area of the JScrollPane. These
 * columns will not scroll horizontally. This is particularly
 * useful when you have one or more columns that identify an object
 * and several additional columns that contain attributes of the
 * object. Placing the identifying columns in these fixed columns
 * ensures that they are always visible while the attributes of the
 * object are scrolled into view.
 * </p>
 * <p>
 * The second JTable contains the remaining columns and is placed in
 * the main viewport and will scroll normally.
 * </p>
 *
 * @author Alex Kluge
 * @version $Revision$, $Date$
 */
public class IdentityTable5 extends JScrollPane
{
    // These names follow a similar pattern to those used in JTable UI delegate.
    public static final String LEFT_ARROW_ACTION_NAME  = "selectPreviousColumn";
    public static final String RIGHT_ARROW_ACTION_NAME = "selectNextColumn";
    public static final String SHIFT_TAB_ACTION_NAME   = "selectPreviousColumnCell";
    public static final String TAB_ACTION_NAME         = "selectNextColumnCell";
    
    private JTable           mainTable;
    private TableColumnModel mainColumnModel      = new DefaultTableColumnModel();
    private JTable           rowHeaderTable;
    private TableColumnModel rowHeaderColumnModel = new DefaultTableColumnModel();
    
    /**
     * Creates a new instance of IdentityTable
     */
    public IdentityTable5(IdentityTableModel baseTableModel)
    {
        // Start out life as an empty scrollpane.
        super();
        
        // Loop over each of the fixed columns, and create columns in the row
        // header column model for them.
        TableColumn column;
        int         columnCount         = baseTableModel.getColumnCount();
        int         currentColumn;
        int         identityColumnCount = baseTableModel.getIdentityColumnCount();
        
        for (currentColumn = 0; currentColumn<identityColumnCount; currentColumn++)
        {
            column = new TableColumn(currentColumn);
            column.setHeaderValue(baseTableModel.getColumnName(currentColumn));
            
            rowHeaderColumnModel.addColumn(column);
        }
        
        // Now loop over the remaining columns, and add them to the main table's column model.
        for (currentColumn = identityColumnCount; currentColumn<columnCount; currentColumn++)
        {
            column = new TableColumn(currentColumn);
            column.setHeaderValue(baseTableModel.getColumnName(currentColumn));
            
            mainColumnModel.addColumn(column);
        }
        
        // Providing a column model here automatically sets the
        // autoCreateColumnsFromModel flag to false.
        rowHeaderTable = new JTable(baseTableModel, rowHeaderColumnModel);
        rowHeaderTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);

        mainTable      = new JTable(baseTableModel, mainColumnModel);
        mainTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        
        // Ensure that both tables have common selections.
        // Save the row selection model for later.
        ListSelectionModel rowSelectionModel = mainTable.getSelectionModel();
        rowHeaderTable.setSelectionModel(rowSelectionModel);
        rowHeaderTable.setPreferredScrollableViewportSize(rowHeaderTable.getPreferredSize());
        
        // This puts the column headers in the upper left corner above the row
        // header columns. This is the same thing that happens when a JTable configures
        // the containing JScrollPane.
        setCorner(JScrollPane.UPPER_LEFT_CORNER, rowHeaderTable.getTableHeader());
        
        JViewport rowHeaderViewPort = new JViewport();
        rowHeaderViewPort.setView(rowHeaderTable);
        setRowHeader(rowHeaderViewPort);

        // Listen for resize events on the row header, and resize the main table as appropriate.
        RowHeaderResizeListener rowHeaderResizeListener = new RowHeaderResizeListener(this,           rowHeaderViewPort,
                                                                                      rowHeaderTable, mainTable);
        rowHeaderTable.addComponentListener(rowHeaderResizeListener);

        setViewportView(mainTable);
        
       installKeybordActions(rowHeaderTable,                      mainTable,
                              rowSelectionModel,                   rowHeaderColumnModel.getSelectionModel(),
                              mainColumnModel.getSelectionModel());
    }
    
    protected void installKeybordActions(JTable             rowHeaderTable,           final JTable       mainTable,
                                         ListSelectionModel rowSelectionModel,        ListSelectionModel headerColumnSelectionModel,
                                         ListSelectionModel mainColumnSelectionModel)
    {
        InputMap rowHeaderInputMap
                    = rowHeaderTable.getInputMap(JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
        ActionMap rowHeaderActionmap = rowHeaderTable.getActionMap();
        ActionMap mainActionMap      = mainTable.getActionMap();


         // Tab action handler, shift forward one cell, with wrapping
        TabKeyActionHandler tabKeyActionHandler
                    = new TabKeyActionHandler(rowHeaderTable,           mainTable,
                                              rowSelectionModel,        headerColumnSelectionModel,
                                              mainColumnSelectionModel, 1,
                                              true);
        // Put the action handler into the appropriate place for both the
        // row header and the main table.
        rowHeaderActionmap.put(TAB_ACTION_NAME, tabKeyActionHandler);
        mainActionMap.put(TAB_ACTION_NAME,      tabKeyActionHandler);

        // Right arrow action handler, shift forward one cell, no wrapping.
        tabKeyActionHandler
                    = new TabKeyActionHandler(rowHeaderTable,           mainTable,
                                              rowSelectionModel,        headerColumnSelectionModel,
                                              mainColumnSelectionModel, 1,
                                              false);

        rowHeaderActionmap.put(RIGHT_ARROW_ACTION_NAME, tabKeyActionHandler);
        mainActionMap.put(RIGHT_ARROW_ACTION_NAME,      tabKeyActionHandler);

        // Shift-Tab action handler, shift backward one cell, with wrapping
        tabKeyActionHandler
                    = new TabKeyActionHandler(rowHeaderTable,           mainTable,
                                              rowSelectionModel,        headerColumnSelectionModel,
                                              mainColumnSelectionModel, -1,
                                              true);

        rowHeaderActionmap.put(SHIFT_TAB_ACTION_NAME, tabKeyActionHandler);
        mainActionMap.put(SHIFT_TAB_ACTION_NAME,      tabKeyActionHandler);

        // Left arrow action handler, shift backward one cell, no wrapping.
        tabKeyActionHandler
                    = new TabKeyActionHandler(rowHeaderTable,           mainTable,
                                              rowSelectionModel,        headerColumnSelectionModel,
                                              mainColumnSelectionModel, -1,
                                              false);

        rowHeaderActionmap.put(LEFT_ARROW_ACTION_NAME, tabKeyActionHandler);
        mainActionMap.put(LEFT_ARROW_ACTION_NAME,      tabKeyActionHandler);
        
        // map the row header page down to the main table's page down.
        Object actionMapKey = rowHeaderInputMap.get(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN, 0));
        rowHeaderActionmap.put(actionMapKey, new RowHeaderActionProxy(mainTable, mainActionMap.get(actionMapKey)));
        
        // and map the page up too
        actionMapKey = rowHeaderInputMap.get(KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP, 0));
        rowHeaderActionmap.put(actionMapKey, new RowHeaderActionProxy(mainTable, mainActionMap.get(actionMapKey)));
    }
    
    /**
     * Return the index of the selected row. Both tables have the same selection
     * model, so the index is consistent.
     * @return An int giving the index of the selected row in the identity table.
     */
    public int getSelectedRow()
    {
        return mainTable.getSelectedRow();
    }
    
    /**
     * Return the row header table.
     * @return A Jtable containing the identity columns from the original table
     *         model.
     */
    public JTable getRowHeaderTable()
    {
        return rowHeaderTable;
    }
    
    /**
     * Fetch the table containing the horizontally scrolling columns.
     * @return A JTable containing the columns from the initial table model
     *         other than the identity columns.
     */
    public JTable getMainTable()
    {
        return mainTable;
    }
    
    /**
     * Sets a default cell renderer for both the row header and the main tables. Be careful
     * as it is also possible to obtain the tables individually and set different default
     * renderers on each.
     *
     * @param columnClass A class, which is matched against the columntype. Columns whose types match
     *                    the class will use the given renderer.
     * @param renderer    A TableCellRenderer used to render cells whose column type matches the given class.
     */
    public void setDefaultRenderer(Class columnClass, TableCellRenderer renderer)
    {
        mainTable.setDefaultRenderer(columnClass, renderer);
        rowHeaderTable.setDefaultRenderer(columnClass, renderer);
    }
    
    /**
     * Return the default cell renderer for the given class from the main table. Be careful
     * as it is also possible to obtain the tables individually and set different default
     * renderers on each.
     */
    public TableCellRenderer getDefaultRenderer(Class columnClass)
    {
        return mainTable.getDefaultRenderer(columnClass);
    }
    
    /**
     * 
     * @return A Table model containing the column model for the main table.
     */
    public TableColumnModel getMainColumnModel()
    {
        return mainColumnModel;
    }

    /**
     * 
     * @return A TableColumnModel containing the column model for the row header.
     */
    public TableColumnModel getRowHeaderColumnModel()
    {
        return rowHeaderColumnModel;
    }
}
