
package com.vizitsolutions.identitytable;

import javax.swing.AbstractAction;
import java.awt.event.ActionEvent;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import java.awt.Rectangle;


/**
 * <p>$Id: $</p>
 *
 * Make tab, shift-tab, and the left and right arrows behave reasonably in the
 * IdentityTable, which is a composite of two separate JTables. Shift the focus
 * to and from the composited table to make it appear that it is all a single
 * table.
 *
 * @author  Alex Kluge
 * @version $Revision: $, $Date: $
 */
public class TabKeyActionHandler extends AbstractAction
{
    /** 
     * The number of cells to shift. This is either +1 for a tab or right arrow,
     * or -1 for a shift tab or left arrow.
     */
    int                delta;
    
    /**
     * The column selection model for the main table. This is used to
     * set the appropriate column as selected when the keyboard is used to
     * navigate to another cell. The selected cell is the intersection of the
     * celected column and the selected row.
     * 
     * @see rowSelectionModel
     */
    ListSelectionModel mainColumnSelectionModel;
    
    /**
     * The main table, contains the table model columns other than the
     * row header columns.
     */
    JTable             mainTable;
    
    /**
     * The column selection model for the row header table. This is used to
     * set the appropriate column as selected when the keyboard is used to
     * navigate to another cell. The selected cell is the intersection of the
     * selected column and the selected row.
     * 
     * @see rowSelectionModel
     */
    ListSelectionModel rowHeaderColumnSelectionModel;
    
    /**
     * This table is the row header for the table.
     */
    JTable             rowHeaderTable;
    
    /**
     * The selected row. The row selection model, and hence the selected row,
     * spans both of the component tables, so the same row is always selected
     * in both tables.
     * 
     * @see mainColumnSelectionModel, rowHeaderColumnSelectionModel
     */
    ListSelectionModel rowSelectionModel;
    
    /**
     * Describes the action when the focus is shifted from the end of a row. If
     * wrap is true, continue on to the next or previous row. If false, then do
     * nothing.
     */
    boolean            wrap;

    /**
     * Construct a TabKeyActionHandler. The TabKeyActionHandler handles the tab
     * as well as the left and right arrow keys. The selected cell is shifted
     * forward or backward as given by the value of delta (+1, -1).
     * 
     * @param rowHeaderTable    A JTable containing the row header columns for the
     *                          identity table.
     * @param mainTable         A JTable containing the remainder of the columns.
     * @param rowSelectionModel 
     * @param rowHeaderColumnSelectionModel
     * @param mainColumnSelectionModel
     * @param delta
     * @param wrap
     */
    public TabKeyActionHandler(JTable             rowHeaderTable,           JTable             mainTable,
                               ListSelectionModel rowSelectionModel,        ListSelectionModel headerColumnSelectionModel,
                               ListSelectionModel mainColumnSelectionModel, int                delta,
                               boolean            wrap)
    {
        this.rowHeaderTable                = rowHeaderTable;
        this.mainTable                     = mainTable;
        this.rowSelectionModel             = rowSelectionModel;
        this.rowHeaderColumnSelectionModel = headerColumnSelectionModel;
        this.mainColumnSelectionModel      = mainColumnSelectionModel;
        this.delta                         = delta;
        this.wrap                          = wrap;
    }
    
    /**
     * Handle the key event by shifting the selected cell forward or backward
     * one column.
     * 
     * @param event
     */
    public void actionPerformed(ActionEvent event)
    {
        JTable currentTable = (JTable)event.getSource();
        
        // If the table is being edited and we can't stop it, return immediately.
        if (currentTable.isEditing() && !currentTable.getCellEditor().stopCellEditing())
        {
            return;
        }
        
        adjustColumn(currentTable);
    }
    
    /**
     * Adjust the selected table cell forward or backward one column.
     * 
     * @param currentTable
     */
    public void adjustColumn(JTable currentTable)
    {
        // This handles the case where no cell is selected. Just as with the
        // standard JTable, the selection is shifted onto the JTable.
        int rowIndex = rowSelectionModel.getLeadSelectionIndex();
        if (rowIndex < 0)
        {
            rowIndex = 0;
        }
        
        // The column of the currently selected cell.
        int columnIndex;
        
        // We have to know which table is the crrently selected table before we
        // can adjust the selected cell. Deal with the rowHeaderTable case
        // first.
        if (currentTable == rowHeaderTable)
        {
            // 
            columnIndex = rowHeaderColumnSelectionModel.getLeadSelectionIndex();
            // If no column is selected
            if (columnIndex < 0)
            {
                // shift to the first column
                columnIndex = 0;
            }
            else
            {
              // Move the cell forward or backward as per the (+1, -1) value of
              // delta.
              columnIndex += delta;
            }

            // If we have walked off the edge of the row header table.
            if (columnIndex > rowHeaderTable.getColumnCount()-1)
            {
                // Stepped from the row header onto the left edge (0 column) of the main table.
                Rectangle currentCell = mainTable.getCellRect(rowIndex, 0, true);
                // Make sure that the cell is visible. This may scroll the main table.
                mainTable.scrollRectToVisible(currentCell);
                // Clear the selection in the row header.
                rowHeaderColumnSelectionModel.clearSelection();
                // Shift the focus from the row header table to the main table.
                mainTable.requestFocusInWindow();
                // We know that the selected column is the first main column.
                mainColumnSelectionModel.setSelectionInterval(0, 0);
                // And we are on the same row as previously.
                rowSelectionModel.setSelectionInterval(rowIndex, rowIndex);
                return;
            }
            else if (columnIndex < 0)
            {
                // We just walked off the left edge of the row. For example,
                // shift-tab in the first column, if we don't wrap to the next
                // column, we're done.
                if (!wrap)
                {
                    return;
                }
                // We will want the last column in the previous row.
                rowIndex -=1;
                columnIndex = mainTable.getColumnCount()-1;
                
                // If we just stepped off the top of the table too
                if (rowIndex < 0)
                {
                    // Wrap around to the bottom, just as the stock JTable.
                    rowIndex = mainTable.getRowCount()-1;
                }
                
                // Wrap from the left edge the header column, to the right edge of the previous
                // column in the main table. Scroll to the appropriate column in the main table.
                Rectangle currentCell = mainTable.getCellRect(rowIndex, columnIndex, true);
                mainTable.scrollRectToVisible(currentCell);
                rowHeaderColumnSelectionModel.clearSelection();
                mainTable.requestFocusInWindow();
                mainColumnSelectionModel.setSelectionInterval(columnIndex, columnIndex);
                rowSelectionModel.setSelectionInterval(rowIndex, rowIndex);
                return;
            }
            else
            {
                // Still within the row header, so select, and scroll to the appropriate column.
                Rectangle currentCell = rowHeaderTable.getCellRect(rowIndex, columnIndex, true);
                rowHeaderTable.scrollRectToVisible(currentCell);
                rowHeaderColumnSelectionModel.setSelectionInterval(columnIndex, columnIndex);
                rowSelectionModel.setSelectionInterval(rowIndex, rowIndex);
                return;
            }
        }
        else
        {
            // A cell in the main table is the currently selected cell.
            columnIndex = mainColumnSelectionModel.getLeadSelectionIndex();
            if (columnIndex < 0)
            {
                columnIndex = 0;
            }
            else
            {
                columnIndex += delta;
            }
            
            // If we have walked off the right edge of the main table.
            if (columnIndex > mainTable.getColumnCount()-1)
            {
                // If we don't wrap to the next column, we're done.
                if (!wrap)
                {
                    return;
                }
                
                // Wrap to the first column of the next row.
                rowIndex += 1;
                columnIndex = 0;
                
                // And if we walked off the bottom of the table,
                if (rowIndex > mainTable.getRowCount()-1)
                {
                    // wrap to the top.
                    rowIndex = 0;
                }
                
                // We have walked of the right edge of the main table. Step around to the zero column of
                // the next row of the row header. Scroll to the selected row header cell, and to the
                // leftmost (0) column of the main table.
                Rectangle currentCell = rowHeaderTable.getCellRect(rowIndex, columnIndex, true);
                rowHeaderTable.scrollRectToVisible(currentCell);
                currentCell = mainTable.getCellRect(0, 0, true);
                mainTable.scrollRectToVisible(currentCell);
                mainColumnSelectionModel.clearSelection();
                rowHeaderTable.requestFocusInWindow();
                rowHeaderColumnSelectionModel.setSelectionInterval(columnIndex, columnIndex);
                rowSelectionModel.setSelectionInterval(rowIndex, rowIndex);
                return;
            }
            else if (columnIndex < 0)
            {
                // We just stepped from the left edge of the main table to the right edge of the header table.
                // Select and scroll to the rightmost (highest value) column in the row header.
                Rectangle currentCell = rowHeaderTable.getCellRect(rowIndex, columnIndex, true);
                rowHeaderTable.scrollRectToVisible(currentCell);
                mainColumnSelectionModel.clearSelection();
                rowHeaderTable.requestFocusInWindow();
                // This is the highest (rightmost) column in the header.
                columnIndex = rowHeaderTable.getColumnCount()-1;
                rowHeaderColumnSelectionModel.setSelectionInterval(columnIndex, columnIndex);
                rowSelectionModel.setSelectionInterval(rowIndex, rowIndex);
                return;
            }
            else
            {
                // Step to the next cell in the main table.
                Rectangle currentCell = mainTable.getCellRect(rowIndex, columnIndex, true);
                mainTable.scrollRectToVisible(currentCell);
                mainColumnSelectionModel.setSelectionInterval(columnIndex, columnIndex);
                rowSelectionModel.setSelectionInterval(rowIndex, rowIndex);
                return;
            }
        }
    }
}
